@@ -2,7 +2,7 @@
 //
 // Author: Mike Kestner <mkestner@speakeasy.net>
 //
-// Copyright (c) 2001-2003 Mike Kestner 
+// Copyright (c) 2001-2003 Mike Kestner
 // Copyright (c) 2003-2005 Novell, Inc.
 // Copyright (c) 2007 Novell, Inc.
 //
@@ -30,24 +30,30 @@ namespace GtkSharp.Generation {
 
 	public class Signal {
 
-		private string name;
-		private XmlElement elem;
-		private ReturnValue retval;
-		private Parameters parms;
-		private ClassBase container_type;
+		bool marshaled;
+		string name;
+		XmlElement elem;
+		ReturnValue retval;
+		Parameters parms;
+		ClassBase container_type;
 
 		public Signal (XmlElement elem, ClassBase container_type)
 		{
 			this.elem = elem;
 			name = elem.GetAttribute ("name");
+			marshaled = elem.GetAttribute ("manual") == "true";
 			retval = new ReturnValue (elem ["return-type"]);
 			parms = new Parameters (elem["parameters"]);
 			this.container_type = container_type;
 		}
 
+		bool Marshaled {
+			get { return marshaled; }
+		}
+
 		public string Name {
 			get {
-				return name; 
+				return name;
 			}
 			set {
 				name = value;
@@ -61,7 +67,7 @@ namespace GtkSharp.Generation {
 				Statistics.ThrottledCount++;
 				return false;
 			}
-			
+
 			if (!parms.Validate () || !retval.Validate ()) {
 				Console.Write (" in signal " + Name + " ");
 				Statistics.ThrottledCount++;
@@ -85,6 +91,31 @@ namespace GtkSharp.Generation {
 			}
 		}
 
+		string CallbackSig {
+			get {
+				string result = "";
+				for (int i = 0; i < parms.Count; i++) {
+					if (i > 0)
+						result += ", ";
+
+					Parameter p = parms [i];
+					if (p.PassAs != "" && !(p.Generatable is StructBase))
+						result += p.PassAs + " ";
+					result += (p.MarshalType + " arg" + i);
+				}
+
+				return result;
+			}
+		}
+
+		string CallbackName {
+			get { return Name + "SignalCallback"; }
+		}
+
+		string DelegateName {
+			get { return Name + "SignalDelegate"; }
+		}
+
                 private string EventArgsName {
                         get {
                                 if (IsEventHandler)
@@ -93,7 +124,7 @@ namespace GtkSharp.Generation {
                                         return Name + "Args";
                         }
                 }
-                                                                                                                        
+
                 private string EventArgsQualifiedName {
                         get {
                                 if (IsEventHandler)
@@ -102,7 +133,7 @@ namespace GtkSharp.Generation {
                                         return container_type.NS + "." + Name + "Args";
                         }
                 }
-                                                                                                                        
+
                 private string EventHandlerName {
                         get {
                                 if (IsEventHandler)
@@ -113,7 +144,7 @@ namespace GtkSharp.Generation {
                                         return Name + "Handler";
                         }
                 }
-                                                                                                                        
+
                 private string EventHandlerQualifiedName {
                         get {
                                 if (IsEventHandler)
@@ -175,6 +206,82 @@ namespace GtkSharp.Generation {
 			}
 		}
 
+		public string GenArgsInitialization (StreamWriter sw)
+		{
+			if (parms.Count > 1)
+				sw.WriteLine("\t\t\t\targs.Args = new object[" + (parms.Count - 1) + "];");
+			string finish = "";
+			for (int idx = 1; idx < parms.Count; idx++) {
+				Parameter p = parms [idx];
+				IGeneratable igen = p.Generatable;
+				if (p.PassAs != "out") {
+					if (igen is ManualGen) {
+						sw.WriteLine("\t\t\t\tif (arg{0} == IntPtr.Zero)", idx);
+						sw.WriteLine("\t\t\t\t\targs.Args[{0}] = null;", idx - 1);
+						sw.WriteLine("\t\t\t\telse {");
+						sw.WriteLine("\t\t\t\t\targs.Args[" + (idx - 1) + "] = " + p.FromNative ("arg" + idx)  + ";");
+						sw.WriteLine("\t\t\t\t}");
+					} else
+						sw.WriteLine("\t\t\t\targs.Args[" + (idx - 1) + "] = " + p.FromNative ("arg" + idx)  + ";");
+				}
+				if (igen is StructBase && p.PassAs == "ref")
+					finish += "\t\t\t\tif (arg" + idx + " != IntPtr.Zero) System.Runtime.InteropServices.Marshal.StructureToPtr (args.Args[" + (idx-1) + "], arg" + idx + ", false);\n";
+				else if (p.PassAs != "")
+					finish += "\t\t\t\targ" + idx + " = " + igen.ToNativeReturn ("((" + p.CSType + ")args.Args[" + (idx - 1) + "])") + ";\n";
+			}
+			return finish;
+		}
+
+		public void GenArgsCleanup (StreamWriter sw, string finish)
+		{
+			if (IsVoid && finish.Length == 0)
+				return;
+
+			sw.WriteLine("\n\t\t\ttry {");
+			sw.Write (finish);
+			if (!IsVoid) {
+				if (retval.CSType == "bool") {
+					sw.WriteLine ("\t\t\t\tif (args.RetVal == null)");
+					sw.WriteLine ("\t\t\t\t\treturn false;");
+				}
+				sw.WriteLine("\t\t\t\treturn " + SymbolTable.Table.ToNativeReturn (retval.CType, "((" + retval.CSType + ")args.RetVal)") + ";");
+			}
+			sw.WriteLine("\t\t\t} catch (Exception) {");
+			sw.WriteLine ("\t\t\t\tException ex = new Exception (\"args.RetVal or 'out' property unset or set to incorrect type in " + EventHandlerQualifiedName + " callback\");");
+			sw.WriteLine("\t\t\t\tGLib.ExceptionManager.RaiseUnhandledException (ex, true);");
+
+			sw.WriteLine ("\t\t\t\t// NOTREACHED: above call doesn't return.");
+			sw.WriteLine ("\t\t\t\tthrow ex;");
+			sw.WriteLine("\t\t\t}");
+		}
+
+		public void GenCallback (StreamWriter sw)
+		{
+			if (IsEventHandler)
+				return;
+
+			sw.WriteLine ("\t\t[UnmanagedFunctionPointer (CallingConvention.Cdecl)]");
+			sw.WriteLine ("\t\tdelegate " + retval.ToNativeType + " " + DelegateName + " (" + CallbackSig + ", IntPtr gch);");
+			sw.WriteLine ();
+			sw.WriteLine ("\t\tstatic " + retval.ToNativeType + " " + CallbackName + " (" + CallbackSig + ", IntPtr gch)");
+			sw.WriteLine("\t\t{");
+			sw.WriteLine("\t\t\t{0} args = new {0} ();", EventArgsQualifiedName);
+			sw.WriteLine("\t\t\ttry {");
+			sw.WriteLine("\t\t\t\tGLib.Signal sig = ((GCHandle) gch).Target as GLib.Signal;");
+			sw.WriteLine("\t\t\t\tif (sig == null)");
+			sw.WriteLine("\t\t\t\t\tthrow new Exception(\"Unknown signal GC handle received \" + gch);");
+			sw.WriteLine();
+			string finish = GenArgsInitialization (sw);
+			sw.WriteLine("\t\t\t\t{0} handler = ({0}) sig.Handler;", EventHandlerQualifiedName);
+			sw.WriteLine("\t\t\t\thandler (GLib.Object.GetObject (arg0), args);");
+			sw.WriteLine("\t\t\t} catch (Exception e) {");
+			sw.WriteLine("\t\t\t\tGLib.ExceptionManager.RaiseUnhandledException (e, false);");
+			sw.WriteLine("\t\t\t}");
+			GenArgsCleanup (sw, finish);
+			sw.WriteLine("\t\t}");
+			sw.WriteLine();
+		}
+
 		private bool NeedNew (ClassBase implementor)
 		{
 			return elem.HasAttribute ("new_flag") ||
@@ -190,7 +297,7 @@ namespace GtkSharp.Generation {
 			string ns = container_type.NS;
 
 			StreamWriter sw = gen_info.OpenStream (EventHandlerName);
-			
+
 			sw.WriteLine ("namespace " + ns + " {");
 			sw.WriteLine ();
 			sw.WriteLine ("\tusing System;");
@@ -203,12 +310,18 @@ namespace GtkSharp.Generation {
 				sw.WriteLine ("\t\tpublic " + parms[i].CSType + " " + parms[i].StudlyName + "{");
 				if (parms[i].PassAs != "out") {
 					sw.WriteLine ("\t\t\tget {");
-					sw.WriteLine ("\t\t\t\treturn (" + parms[i].CSType + ") Args[" + (i - 1) + "];");
+					if (SymbolTable.Table.IsInterface (parms [i].CType))
+						sw.WriteLine ("\t\t\t\treturn {0}Adapter.GetObject (Args [{1}] as GLib.Object);", parms [i].CSType, i - 1);
+					else
+						sw.WriteLine ("\t\t\t\treturn ({0}) Args [{1}];", parms [i].CSType, i - 1);
 					sw.WriteLine ("\t\t\t}");
 				}
 				if (parms[i].PassAs != "") {
 					sw.WriteLine ("\t\t\tset {");
-					sw.WriteLine ("\t\t\t\tArgs[" + (i - 1) + "] = (" + parms[i].CSType + ")value;");
+					if (SymbolTable.Table.IsInterface (parms [i].CType))
+						sw.WriteLine ("\t\t\t\tArgs [{0}] = value is {1}Adapter ? (value as {1}Adapter).Implementor : value;", i - 1, parms [i].CSType);
+					else
+						sw.WriteLine ("\t\t\t\tArgs[" + (i - 1) + "] = (" + parms[i].CSType + ")value;");
 					sw.WriteLine ("\t\t\t}");
 				}
 				sw.WriteLine ("\t\t}");
@@ -252,7 +365,7 @@ namespace GtkSharp.Generation {
 					string call_parm = p.CallString;
 
 					if (p.IsUserData && parms.IsHidden (p) && !parms.HideData && (i == 1 || parms [i - 1].Scope != "notified")) {
-						call_parm = "IntPtr.Zero"; 
+						call_parm = "IntPtr.Zero";
 					}
 
 					result += ", " + call_parm;
@@ -306,7 +419,7 @@ namespace GtkSharp.Generation {
 			glue.WriteLine ("}");
 
 			StreamWriter sw = gen_info.Writer;
-			sw.WriteLine ("\t\t[DllImport (\"{0}\")]", gen_info.GluelibName);
+			sw.WriteLine ("\t\t[DllImport (\"{0}\", CallingConvention = CallingConvention.Cdecl)]", gen_info.GluelibName);
 			sw.WriteLine ("\t\tstatic extern {0} {1} ({2});\n", retval.MarshalType, glue_name, parms.ImportSignature);
 			GenVMDeclaration (sw, null);
 			sw.WriteLine ("\t\t{");
@@ -381,13 +494,13 @@ namespace GtkSharp.Generation {
 			bool use_glue = gen_info.GlueEnabled && implementor == null && ClassFieldName.Length > 0;
 			string glue_name = String.Empty;
 			ManagedCallString call = new ManagedCallString (parms, true);
-			sw.WriteLine ("\t\t[GLib.CDeclCallback]");
+			sw.WriteLine ("\t\t[UnmanagedFunctionPointer (CallingConvention.Cdecl)]");
 			sw.WriteLine ("\t\tdelegate " + retval.ToNativeType + " " + Name + "VMDelegate (" + parms.ImportSignature + ");\n");
 
 			if (use_glue) {
 				glue = gen_info.GlueWriter;
 				glue_name = String.Format ("{0}sharp_{1}_override_{2}", container_type.NS.ToLower ().Replace (".", "_"), container_type.Name.ToLower (), ClassFieldName);
-				sw.WriteLine ("\t\t[DllImport (\"{0}\")]", gen_info.GluelibName);
+				sw.WriteLine ("\t\t[DllImport (\"{0}\", CallingConvention = CallingConvention.Cdecl)]", gen_info.GluelibName);
 				sw.WriteLine ("\t\tstatic extern void {0} (IntPtr gtype, {1}VMDelegate cb);\n", glue_name, Name);
 				glue.WriteLine ("void {0} (GType gtype, gpointer cb);\n", glue_name);
 				glue.WriteLine ("void\n{0} (GType gtype, gpointer cb)", glue_name);
@@ -437,6 +550,11 @@ namespace GtkSharp.Generation {
 		{
 			string args_type = IsEventHandler ? "" : ", typeof (" + EventArgsQualifiedName + ")";
 
+			if (Marshaled) {
+				GenCallback (sw);
+				args_type = ", new " + DelegateName + "(" + CallbackName + ")";
+			}
+
 			sw.WriteLine("\t\t[GLib.Signal("+ CName + ")]");
 			sw.Write("\t\tpublic ");
 			if (NeedNew (implementor))
@@ -467,9 +585,8 @@ namespace GtkSharp.Generation {
 			else
 				GenChainVirtualMethod (sw, implementor);
 			GenEvent (sw, implementor, "this");
-			
+
 			Statistics.SignalCount++;
 		}
 	}
 }
-
