@@ -49,14 +49,21 @@ namespace GLib {
 		
 		internal static Hashtable source_handlers = new Hashtable ();
 		
-		[DllImport("libglib-2.0-0.dll")]
+		[DllImport("libglib-2.0-0.dll", CallingConvention=CallingConvention.Cdecl)]
 		static extern bool g_source_remove (uint tag);
 
 		public static bool Remove (uint tag)
 		{
-			lock (Source.source_handlers)
-				source_handlers.Remove (tag);
-			return g_source_remove (tag);
+			// g_source_remove always returns true, so we follow that
+			bool ret = true;
+
+			lock (Source.source_handlers) {
+				if (source_handlers.Contains (tag)) {
+					source_handlers.Remove (tag);
+					ret = g_source_remove (tag);
+				}
+			}
+			return ret;
 		}
 	}
 }
